#include <sta/math/algorithms/kalmanFilter.hpp>
#include <sta/math/linalg/linalg.hpp>
#include <sta/debug/debug.hpp>
#include <sta/debug/assert.hpp>
#include <sta/debug/profile.hpp>

namespace sta
{

namespace math
{

KalmanFilter::KalmanFilter(matrix F, matrix B, matrix H, matrix Q, matrix R) : F_{F}, B_{B}, H_{H}, Q_{Q}, R_{R}, n_{F.get_cols()}
{
	STA_ASSERT_MSG(F.get_rows() == B.get_rows(), "#rows mismatch: F, B!");
	STA_ASSERT_MSG(F.get_cols() == H.get_cols(), "#cols mismatch: F, H!");
	STA_ASSERT_MSG(F.get_rows() == Q.get_rows(), "#rows mismatch: F, Q!");
	STA_ASSERT_MSG(Q.get_cols() == Q.get_rows(), "Q not square!");
    STA_ASSERT_MSG(F.get_cols() == F.get_rows(), "F not square!");
	STA_ASSERT_MSG(H.get_rows() == R.get_rows(), "#rows mismatch: H, R");
	STA_ASSERT_MSG(R.get_rows() == R.get_cols(), "#R not square");
    identity_ = matrix::eye(n_);
}

KalmanFilter::~KalmanFilter()
{
    // Destructor implementation
}

KalmanState KalmanFilter::predict(KalmanState state, matrix u)
{
    // Predict step implementation
    // Update the state based on the system dynamics
	state.x = (F_ * state.x) + (B_ * u);
    // Update the error covariance matrix
    state.error = F_ *( state.error * F_.T()) + Q_;
    return state;
}

KalmanState KalmanFilter::predict(KalmanState state)
{
	state.x = F_ * state.x;

	/*
	state.x.set(0, 0, state.x[0] + F_.get_idx(0, 3) * state.x[3] + F_.get_idx(0, 6) * state.x[6]);
	state.x.set(1, 0, state.x[1] + F_.get_idx(1, 4) * state.x[4] + F_.get_idx(1, 7) * state.x[7]);
	state.x.set(2, 0, state.x[2] + F_.get_idx(2, 5) * state.x[5] + F_.get_idx(2, 8) * state.x[6]);
	state.x.set(3, 0, state.x[3] + F_.get_idx(3, 6) * state.x[6]);
	state.x.set(4, 0, state.x[4] + F_.get_idx(4, 7) * state.x[7]);
	state.x.set(5, 0, state.x[5] + F_.get_idx(5, 8) * state.x[8]);
	state.x.set(6, 0, state.x[6]);
	state.x.set(7, 0, state.x[7]);
	state.x.set(8, 0, state.x[8]);
	*/

	state.error = F_ * (state.error * F_.T()) + Q_;

	return state;
}

KalmanState KalmanFilter::correct(KalmanState state, matrix z)
{
    // Correct step implementation
    // Calculate the Kalman gain
    matrix K = state.error * H_.T() * linalg::inv(H_ * state.error * H_.T() + R_);
    // Update the state based on the measurement
    state.x = state.x + K * (z - H_ * state.x); //TODO check transpose

    // Update the error covariance matrix
    state.error = (identity_ - K * H_) * state.error;
    return state;
}

KalmanState KalmanFilter::correct(KalmanState state, matrix z, matrix H, matrix R)
{
    // Correct step implementation
    // Calculate the Kalman gain
    matrix K = state.error * H.T() * linalg::inv(H * state.error * H.T() + R);
    // Update the state based on the measurement
    state.x = state.x + K * (z - H * state.x); //TODO check transpose
    // Update the error covariance matrix
    //state.error = (matrix::eye(state.x.get_rows()) - K * H) * state.error;
    state.error = (identity_ - K * H) * state.error;
    return state;
}

} // namespace math

} // namespace sta
